/*

    This file is part of the Maude 3 interpreter.

    Copyright 1997-2003 SRI International, Menlo Park, CA 94025, USA.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

*/

//
//	Class for dynamic string tables
//
#ifndef _stringTable_hh_
#define _stringTable_hh_

class StringTable
{
public:
  StringTable(int size = 1024);
  ~StringTable();
  int encode(const char* name);
  const char* name(int code) const;

private:
  static int hash(const char* name);
  static int hash2(const char* name);
  void resize();

  Vector<char*> stringTable;
  Vector<int> hashTable;
};

inline const char*
StringTable::name(int code) const
{
  return stringTable[code];
}

#endif
